import pathlib

import polars as pl
import pytest

from maplib import (
    Model,
    Template,
    IRI,
    Prefix,
    Triple,
    Variable,
    Parameter,
    Literal,
    xsd,
    RDFType, rdf,
)
from polars.testing import assert_frame_equal
PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"

def test_construct_with_blank():
    m = Model()
    m.read(TESTDATA_PATH / "read_ntriples.nt")
    res = m.query("""
    PREFIX :<urn:maplib:ex> 
    CONSTRUCT {
        ?a :pred [a :bdc ].
    } WHERE {
        ?a a ?b .
    }
    """)
    assert len(res) == 2
    for r in res:
        assert r.height == 2

def test_insert_with_blank():
    m = Model()
    m.read(TESTDATA_PATH / "read_ntriples.nt")
    m.update("""
    PREFIX :<urn:maplib:ex> 
    INSERT {
        ?a :pred [a :bdc ].
    } WHERE {
        ?a a ?b .
    }
    """)

    df = m.query("""
    SELECT ?a ?b ?c WHERE {?a ?b ?c}
    """)
    assert df.height == 12



import pytest

from maplib import Model, MaplibException
import polars as pl
import pathlib

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata" / "jsonlds"

def test_read_jsonld():
    m = Model()
    with pytest.raises(MaplibException) as e:
        m.read(TESTDATA_PATH / "test.jsonld")

def test_csvw_example():
    CSVW_PATH = TESTDATA_PATH / "csvw.jsonld"
    with open(CSVW_PATH) as f:
        csvw = f.read()
    m = Model()
    m.read(TESTDATA_PATH / "grit_bins.json", known_contexts={"http://www.w3.org/ns/csvw":csvw})
    df = m.query("SELECT * WHERE {?a ?b ?c}")
    assert df.height == 17


import os
from typing import List, Optional

import pytest

from .disk import disk_params
from maplib import Model
from polars.testing import assert_frame_equal
import polars as pl
import pathlib
import time

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata" / "jsons"

@pytest.mark.parametrize("disk", disk_params())
def test_map_json_1(disk):
    json_1 = TESTDATA_PATH / "1.json"
    m = Model(storage_folder=disk)
    start_map = time.time()
    m.map_json(str(json_1))
    map_took = time.time() - start_map
    print(f"Map took {round(map_took, 5)}")
    start_query = time.time()
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 18
    query_took = time.time() - start_query
    print(f"Query took {round(query_took, 5)}")

    print(m.writes("turtle"))
    df2 = m.query("""
    PREFIX fx:  <http://sparql.xyz/facade-x/ns/> 
    PREFIX xyz: <http://sparql.xyz/facade-x/data/>
    SELECT ?title WHERE {
    ?n1 a fx:root .
    ?n1 xyz:glossary ?n2 .
    ?n2 xyz:title ?title . 
    }""")
    expect2 = pl.from_repr("""
┌──────────────────┐
│ title            │
│ ---              │
│ str              │
╞══════════════════╡
│ example glossary │
└──────────────────┘
""")
    assert_frame_equal(df2, expect2)

@pytest.mark.parametrize("disk", disk_params())
def test_map_json_2(disk):
    json_2 = TESTDATA_PATH / "2.json"
    m = Model(storage_folder=disk)
    m.map_json(str(json_2))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 56
    # We expect this number to increase when we start caring about array ordering
    df2 = m.query("""
    PREFIX fx:  <http://sparql.xyz/facade-x/ns/> 
    PREFIX xyz: <http://sparql.xyz/facade-x/data/>
    PREFIX mj:  <urn:maplib_json:> 
    SELECT ?item WHERE {
    ?n1 a fx:root .
    ?n1 xyz:menu ?n2 .
    ?n2 xyz:items ?arr .
    ?arr ?ref ?item .
    } """)
    assert df2.height == 22

def test_map_json_3():
    json_3 = TESTDATA_PATH / "3.json"
    m = Model()
    m.map_json(str(json_3))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 88
    # We expect this number to increase when we start caring about array ordering

    df2 = m.query("""
    PREFIX fx:  <http://sparql.xyz/facade-x/ns/> 
    PREFIX xyz: <http://sparql.xyz/facade-x/data/>
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#> 
    PREFIX mj:  <urn:maplib_json:> 
    SELECT ?log ?betaserver ?myfloat WHERE {
    ?n1 a fx:root .
    ?n1 xyz:web-app ?n2 .
    ?n2 xyz:servlet ?arr .
    ?arr rdf:_5 ?n3 .
    ?n3 xyz:init-param ?n4 . 
    ?n4 xyz:log ?log .
    ?n4 xyz:betaServer ?betaserver .
    ?n5 xyz:myFloat ?myfloat .
    }""")
    assert df2.height == 1
    assert df2.get_column("log").dtype == pl.Int64
    assert df2.get_column("betaserver").dtype == pl.Boolean
    assert df2.get_column("myfloat").dtype == pl.Float64

def test_map_json_string():
    s = """
    { "maplib_users": {
    "Alice": { "name": "Alice", "address": { "city": "Swansea" } },
    "Bob": { "name": "Bob", "address": { "city": "Swansea" } }
    }}
    """
    m = Model()
    m.map_json(s)
    df_before = m.query("SELECT * WHERE {?a ?b ?c}")
    r = m.writes(format="turtle", prefixes={
        "mj":"urn:maplib_json:"
    })
    print(r)
    m2 = Model()
    m2.reads(r, format="turtle")
    df_after = m2.query("SELECT * WHERE {?a ?b ?c}")

    assert df_before.height == df_after.height

def test_map_json_list_string():
    s = """
    { "maplib_users": [
    { "name": "Alice", "address": { "city": "Swansea" } },
    { "name": "Bob", "address": { "city": "Swansea" } }
    ]}
    """
    m = Model()
    m.map_json(s)
    height = m.query("SELECT * WHERE {?a ?b ?c}").height
    r = m.writes(format="turtle", prefixes={
        "mj":"urn:maplib_json:"
    })
    print(r)
    m2 = Model()
    m2.reads(r, format="turtle")
    height_after = m2.query("SELECT * WHERE {?a ?b ?c}").height
    assert height == height_after

@pytest.mark.parametrize("disk", disk_params())
def test_insert(disk):
    s = """
    { "maplib_users": [
    { "name": "Alice", "address": { "city": "Swansea" } },
    { "name": "Bob", "address": { "city": "Swansea" } }
    ]}
    """
    m = Model(storage_folder=disk)
    m.map_json(s, graph="urn:graph:tmp")
    users = """
        PREFIX :<https://github.com/DataTreehouse/maplib/users#>
        CONSTRUCT { ?u a :SwanseaUser; :name ?name } 
        WHERE {
            ?root xyz:maplib_users ?users.
            ?users ?i ?u.
            ?u xyz:address/xyz:city "Swansea"; xyz:name ?name.
        }"""
    m.insert(users, source_graph="urn:graph:tmp")
    #print(m.writes(format="turtle", prefixes={"":"https://github.com/DataTreehouse/maplib/users#"}))
    df = m.query("SELECT * WHERE {?a ?b ?c}")
    assert df.height == 4


# r'''
# # Overview
#
# '''

import logging
logger = logging.getLogger(__name__)

__all__ = [
    "Model",
    "a",
    "Triple",
    "SolutionMappings",
    "IndexingOptions",
    "ValidationReport",
    "Instance",
    "Template",
    "Argument",
    "Parameter",
    "Variable",
    "RDFType",
    "xsd",
    "rdf",
    "rdfs",
    "owl",
    "IRI",
    "Literal",
    "Prefix",
    "BlankNode",
    "explore",
    "add_triples",
    "generate_templates",
    "MaplibException",
]

import pathlib
from importlib.metadata import version
from .maplib import *
from .adding_triples import add_triples
from .template_generator import generate_templates

"""
http://www.w3.org/1999/02/22-rdf-syntax-ns#type
"""
a = rdf.type

if (pathlib.Path(__file__).parent.resolve() / "graph_explorer").exists():
    from .graph_explorer import explore as _explore
else:

    def _explore(
        m: "Model",
        host: str = "localhost",
        port: int = 8000,
        bind: str = "localhost",
        popup=True,
        fts=True,
        fts_path:str="fts",
    ):
        print("Contact Data Treehouse to try!")


def explore(*args, **kwargs):
    """Deprecated way to start an explore session.
Use the explore method on a Model object instead
"""
    logger.warn("Calling `maplib.explore` is deprecated, use `m.explore()` on a `Model` object instead")
    if kwargs.get("popup") == None or kwargs.get("popup") == True:
        logger.warn("""Calling explore without a popup argument defaults to it being on.
The popup argument is deprecated, so if you are relying on explore() opening a browser window
please change this to something like

```
import webbrowser
from maplib import Model

m = Model()
...
s = m.explore()
webbrowser.open(s.url, new=2)
```
""")
        kwargs["popup"] = True
    elif kwargs.get("popup") == False:
        logger.warn("The new explore function on a Model, no longer defaults to popping up the browser ")


    return _explore(*args, **kwargs)

__version__ = version("maplib")



from polars.testing import assert_frame_equal
import polars as pl
from maplib import Model
import pathlib
import pytest

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata" / "compressed_iri_issue"

# This test checks an issue with sorting compressed IRIs.
# The naive sorting by prefix first and then by suffix does not work in general.
# See implementation of Ord for in Cats Compressed IRI.
@pytest.mark.parametrize("streaming", [True, False])
def test_compressed_iri_issue(streaming):
    m = Model()
    shape_graph = "urn:eu:shacl"
    m.read(TESTDATA_PATH / "skos-era-skos-SubCategories.ttl", graph=shape_graph)
    m.read(TESTDATA_PATH / "skos-era-skos-SubCategories.ttl", graph=shape_graph)
    m.read(TESTDATA_PATH / "skos-era-skos-SubCategories.ttl", graph=shape_graph)
    m.read(TESTDATA_PATH / "skos-era-skos-SubCategories.ttl", graph=shape_graph)
    m.read(TESTDATA_PATH / "skos-era-skos-SubCategories.ttl", graph=shape_graph)




from typing import List, Optional
import os
import pathlib

PATH_HERE = pathlib.Path(__file__).parent

def disk_params() -> List[Optional[str]]:
    import dotenv
    dotenv.load_dotenv(PATH_HERE.parent.parent/".env")
    if "MAPLIB_EE" in os.environ:
        is_maplib_ee = os.environ["MAPLIB_EE"]
    else:
        is_maplib_ee = False
    if is_maplib_ee:
        return [None, "disk"]
    else:
        return [None]


import pytest
from maplib import Model
from maplib import MaplibException


def test_model_exception():
    m = Model()

    with pytest.raises(MaplibException) as e:
        m.reads("abc", format="turtle", graph="http://example.com/data")

def test_query_groupby_exception():
    m = Model()
    with pytest.raises(MaplibException) as e:
        m.query("""
            SELECT (COUNT(?a) as ?ca) WHERE {?a ?b ?c .} GROUP BY ?ca
        """)

def test_query_filter_exception():
    m = Model()
    with pytest.raises(MaplibException) as e:
        m.query("""
            SELECT * WHERE {
                ?a ?b ?c .
                FILTER("a")
            }
        """)



import polars as pl
import pytest
import rdflib
from maplib.maplib import IndexingOptions
from polars.testing import assert_frame_equal
import pathlib
from maplib import Model

pl.Config.set_fmt_str_lengths(300)

PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"

def test_write_turtle_default_prefixes():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    out = m.writes(format="turtle")
    m2 = Model()
    m2.reads(out, format="turtle")

def test_write_turtle_newline_bug():
    m = Model()
    m.read(str(TESTDATA_PATH / "write_turtle_newlines_bug.nt"))
    out = m.writes(format="turtle")
    m2 = Model()
    m2.reads(out, format="turtle")

def test_write_turtle_provided_prefixes():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    out = m.writes(format="turtle", prefixes={"myfoaf": "http://xmlns.com/foaf/0.1/"})
    m2 = Model()
    m2.reads(out, format="turtle")
    assert "myfoaf:" in out

def test_write_turtle_global_provided_prefixes():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.add_prefixes({"myfoaf": "http://xmlns.com/foaf/0.1/"})
    out = m.writes(format="turtle")
    m2 = Model()
    m2.reads(out, format="turtle")
    assert "myfoaf:" in out

def test_write_turtle_global_provided_prefixes_and_provided():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.add_prefixes({"myfoaf": "http://xmlns.com/foaf/0.1/"})
    out = m.writes(format="turtle", prefixes={})
    m2 = Model()
    m2.reads(out, format="turtle")
    assert "myfoaf:" in out

def test_write_multi_turtle_provided_prefixes():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 24

    out = m.writes(format="turtle", prefixes={"myfoaf": "http://xmlns.com/foaf/0.1/"})
    m2 = Model()
    m2.reads(out, format="turtle")
    df2 = m2.query("""SELECT * WHERE {?a ?b ?c}""")
    expected_path = TESTDATA_PATH / "pretty_turtle_write_multi_provided_expected.csv"
    #df2.write_csv(expected_path)
    df2_expected = pl.read_csv(expected_path)
    df2 = df2.with_columns(
        pl.when(pl.col("a").str.starts_with("_:")).then(pl.lit("blank!")).otherwise(pl.col("a")).alias("a"),
        pl.when(pl.col("c").str.starts_with("_:")).then(pl.lit("blank!")).otherwise(pl.col("c")).alias("c"),
    ).sort(["a", "b", "c"])
    df2_expected = df2_expected.with_columns(
        pl.when(pl.col("a").str.starts_with("_:")).then(pl.lit("blank!")).otherwise(pl.col("a")).alias("a"),
        pl.when(pl.col("c").str.starts_with("_:")).then(pl.lit("blank!")).otherwise(pl.col("c")).alias("c"),
    ).sort(["a", "b", "c"])
    assert_frame_equal(df2_expected, df2)
    assert df2.height == 24


@pytest.mark.skip("This test works but takes too long")
def test_write_multi_turtle_provided_prefixes_stress():
    m = Model(IndexingOptions(subject_object_index=True))
    for i in range(3000):
        m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 24000

    out = m.writes(format="turtle", prefixes={"myfoaf": "http://xmlns.com/foaf/0.1/"})
    m2 = Model()
    m2.reads(out, format="turtle", parallel=True)
    df2 = m2.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df2.height == 24000


def test_write_multi_turtle_provided_prefixes_some_stress():
    m = Model(IndexingOptions(subject_object_index=True))
    for i in range(100):
        m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 800

    out = m.writes(format="turtle", prefixes={"myfoaf": "http://xmlns.com/foaf/0.1/"})
    m2 = Model()
    m2.reads(out, format="turtle", parallel=True)
    df2 = m2.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df2.height == 800


def test_write_lists():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_lists.ttl"))
    df = m.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df.height == 30

    out = m.writes(format="turtle", prefixes={"myfoaf": "http://xmlns.com/foaf/0.1/"})
    print(out)
    m2 = Model()
    m2.reads(out, format="turtle")
    df2 = m2.query("""SELECT * WHERE {?a ?b ?c}""")
    assert df2.height == 30


import polars as pl
import pytest
import rdflib
from polars.testing import assert_frame_equal
import pathlib
from maplib import Model

pl.Config.set_fmt_str_lengths(300)


PATH_HERE = pathlib.Path(__file__).parent
TESTDATA_PATH = PATH_HERE / "testdata"



def test_read_ntriples_utf8_bom():
    # Produced this way:
    #with open(TESTDATA_PATH / "read_ntriples.nt") as f:
    #    c = f.read()
    #with open(TESTDATA_PATH / "read_ntriples_utf8_bom.nt", "w", encoding="utf-8-sig") as f:
    #    f.write(c)

    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples_utf8_bom.nt"))
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?s ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?s ?v ?o
            """
    ).sort(["s", "v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).collect()
    pl.testing.assert_frame_equal(res, expected_df)

def test_read_ntriples():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?s ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?s ?v ?o
            """
    ).sort(["s", "v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).collect()
    pl.testing.assert_frame_equal(res, expected_df)

def test_read_many_different_numbers():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_nums.ttl"))
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>
            PREFIX :<https://example.net/nums/> 

            SELECT ?s ?byte ?ubyte ?short ?ushort WHERE {
            ?s :byte ?byte ;
               :unsignedByte ?ubyte ;
               :short ?short ;
               :unsignedShort ?ushort .
            } 
        """
    )
    assert res.height == 1
    assert res.get_column("byte").dtype == pl.Int8
    assert res.get_column("ubyte").dtype == pl.UInt8
    assert res.get_column("short").dtype == pl.Int16
    assert res.get_column("ushort").dtype == pl.UInt16



def test_read_ntriples_twice_with_replace():
    m = Model()
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"))
    m.read(str(TESTDATA_PATH / "read_ntriples.nt"), replace_graph=True)
    res = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?s ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?s ?v ?o
            """
    ).sort(["s", "v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples.csv"
    # res.write_csv(str(filename))
    assert res.shape == (8, 3)


def test_read_write_ntriples_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="ntriples")
    m2 = Model()
    m2.reads(out_str, format="ntriples")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_write_turtle_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="turtle")
    m2 = Model()
    m2.reads(out_str, format="turtle")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_write_xml_string():
    m = Model()
    with open(TESTDATA_PATH / "read_ntriples.nt") as f:
        ntstring = f.read()
    m.reads(ntstring, format="ntriples")
    out_str = m.writes(format="rdf/xml")
    m2 = Model()
    m2.reads(out_str, format="rdf/xml")
    res = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT ?v ?o WHERE {
            ?s ?v ?o .
            } ORDER BY ?v ?o
            """
    ).sort(["v", "o"])
    # TODO: Fix multitype sorting
    filename = TESTDATA_PATH / "read_ntriples2.csv"
    # res.write_csv(str(filename))
    expected_df = pl.scan_csv(filename).select(["v", "o"]).sort(["v", "o"]).collect()
    pl.testing.assert_frame_equal(res, expected_df)


def test_read_singleton_blanks_turtle_string():
    m = Model()
    m.reads("""
    @prefix ex:<urn:maplib:>.
    _:b a _:dog .
    _:c a _:cat .
    _:cat ex:pred "cat_label" . 
    _:d ex:pred2 "abc" .
    ex:cat ex:pred3 _:e .
    ex:dog a _:f .
    """,
    format="turtle")
    blanks1 = m.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT DISTINCT ?blank WHERE {
                {
                SELECT ?blank WHERE {
                    ?blank ?p ?o .
                    FILTER(ISBLANK(?blank))                
                } 
                }
                UNION
                {
                SELECT ?blank WHERE {
                    ?s ?p ?blank .
                    FILTER(ISBLANK(?blank))                
                }
                } 
            } 
            """
    )
    out_str = m.writes(format="turtle")
    print(out_str)
    m2 = Model()
    m2.reads(out_str, format="turtle")
    blanks2 = m2.query(
        """
            PREFIX foaf:<http://xmlns.com/foaf/0.1/>

            SELECT DISTINCT ?blank WHERE {
                {
                    SELECT ?blank WHERE {
                        ?blank ?p ?o .
                        FILTER(ISBLANK(?blank))                
                    } 
                }
                UNION
                {
                    SELECT ?blank WHERE {
                        ?s ?p ?blank .
                        FILTER(ISBLANK(?blank))                
                    }
                } 
            } 
            """
    )
    assert blanks1.height == blanks2.height

